/*
  ggcap - main.c
  Copyright (c) 2009-2010 Arkadiusz Bokowy

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  If you want to read full version of the GNU General Public License
  see <http://www.gnu.org/licenses/>.

  ** Note: **
  For contact information and the latest version of this program see
  my webpage <http://arkq.awardspace.us/#ggcap>.

*/

#include <string.h>
#include <ctype.h>
#include <getopt.h>
#include <errno.h>

#include <signal.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>

#include "ggcap.h"

void debug(char *msg){printf(msg); fflush(stdout);}

// Used by sigaction to interrupt loop mode
static void stop_loop_mode(int sig){gg_dt.fl.main_loop = 0;}

int main(int argc, char *argv[])
{
	int opt, status;
	uint32_t netp, maskp; //net & subnet mask
	int pcap_fd;
	fd_set rd_fds;
	struct termios old_opts, new_opts;
	struct sigaction sigacts;
	struct bpf_program fprog; //hold compiled program
	struct option longopts[] = {
		{"help", no_argument, NULL, 'h'},
		{"version", no_argument, NULL, 'V'},
		{"iface", required_argument, NULL, 'i'},
		{"read", required_argument, NULL, 'R'},
		{"filter", required_argument, NULL, 'f'},
		{"aliases", required_argument, NULL, 'a'},
		{"verbose", no_argument, NULL, 'v'},
		{"nostatus", no_argument, NULL, 1},
		{"nolocal", no_argument, NULL, 2},
		{"vismod-off", no_argument, NULL, 3},
		{"show-conn-onexit", no_argument, NULL, 4},
		{0, 0, 0, 0}
	};
	char *cp, errbuf[PCAP_ERRBUF_SIZE], errmsg[PCAP_ERRBUF_SIZE + 100];
	register char *cmdbuf = NULL;

	// clear global structure
	bzero(&gg_dt, sizeof(gg_dt));

	// set compiled in defaults
	strcpy(gg_dt.iface, "eth0");
	gg_dt.fl.visualization = 1;

	// arguments parser
	while((opt = getopt_long(argc, argv, "hVi:R:a:fv", longopts, NULL)) != -1)
		switch(opt) {
		case 'h': usage_help(); exit(EXIT_SUCCESS);
		case 'V': printf("%s\n", APP_VER); exit(EXIT_SUCCESS);
		case 'i':
			strncpy(gg_dt.iface, optarg, sizeof(gg_dt.iface) - 1);
			break;
		case 'R':
			strncpy(gg_dt.read_fname, optarg, sizeof(gg_dt.read_fname) - 1);
			break;
		case 'f':
			cmdbuf = optarg; break;
		case 'a':
			if(load_uinalias_table(optarg) == -1)
				fprintf(stderr, "warning: alias-table not loaded: %s\n",
						strerror(errno));
			break;
		case 'v':
			gg_dt.fl.verbose = 1; break;
		case 1:
			gg_dt.fl.drop_status = 1; break;
		case 2:
			gg_dt.fl.drop_local = 1; break;
		case 3:
			gg_dt.fl.visualization = 0; break;
		case 4:
			gg_dt.fl.show_conn_onexit = 1; break;
		default:
			printf("Try '%s --help' for more information.\n", APP_NAME);
			exit(EXIT_FAILURE);
		}

	// prepare GG networks masks
	gg_dt.net = inet_addr(GG_NET);
	gg_dt.netmask = inet_addr(GG_NETMASK);

	traficdb_initialize();
	tcpipdata_collector_initialize();
	spooftool_initialize();

	if(gg_dt.read_fname[0] != 0) { //OFF LINE sniffing mode
		int dlt;
		const char *dlt_name;

		if((gg_dt.pcap = pcap_open_offline(gg_dt.read_fname, errbuf)) == NULL)
			exit_error(errbuf);

		// get link-type information
		dlt = pcap_datalink(gg_dt.pcap);
		dlt_name = pcap_datalink_val_to_name(dlt);

		if(dlt_name == NULL)
			fprintf(stderr, "reading from file %s, link-type %u\n\n",
					gg_dt.read_fname, dlt);
		else
			fprintf(stderr, "reading from file %s, link-type %s (%s)\n\n",
					gg_dt.read_fname, dlt_name, pcap_datalink_val_to_description(dlt));
	}
	else { //ON LINE sniffing mode
		// create and configure pcap handler
		if((gg_dt.pcap = pcap_create(gg_dt.iface, errbuf)) == NULL)
			exit_error(errbuf);
		if((status = pcap_set_snaplen(gg_dt.pcap, BUFSIZ)) != 0){
			sprintf(errmsg, "pcap_set_snaplen failed: %s", pcap_statustostr(status));
			exit_error(errmsg);}
		if((status = pcap_set_promisc(gg_dt.pcap, 1)) != 0){
			sprintf(errmsg, "pcap_set_promisc failed: %s", pcap_statustostr(status));
			exit_error(errmsg);}
		if((status = pcap_set_timeout(gg_dt.pcap, 1000)) != 0){
			sprintf(errmsg, "pcap_set_timeout failed: %s", pcap_statustostr(status));
			exit_error(errmsg);}

		status = pcap_activate(gg_dt.pcap);
		if(status < 0) { //Error!
			cp = pcap_geterr(gg_dt.pcap);
			if(status == PCAP_ERROR)
				exit_error(cp);
			else if((status == PCAP_ERROR_NO_SUCH_DEVICE ||
					status == PCAP_ERROR_PERM_DENIED) && *cp != '\0'){
				sprintf(errmsg, "%s: %s\n(%s)", gg_dt.iface,
						pcap_statustostr(status), cp);
				exit_error(errmsg);}
			else{
				sprintf(errmsg, "%s: %s", gg_dt.iface, pcap_statustostr(status));
				exit_error(errmsg);}
		}
		else if(status > 0) { //OK but... something is wrong
			fprintf(stderr, "warning: %s\n", pcap_geterr(gg_dt.pcap));
		}

		fflush(stderr);
	}

	// get network number and mask
	if(pcap_lookupnet(gg_dt.iface, &netp, &maskp, errbuf) < 0){
		netp = 0; maskp = 0;
		fprintf(stderr, "warning: %s\n", errbuf);}

	// get captured (local) device IP
	gg_dt.localip = get_devip(gg_dt.iface);

	if(pcap_compile(gg_dt.pcap, &fprog, cmdbuf, 0, maskp) < 0)
		exit_error(pcap_geterr(gg_dt.pcap));

	// catch singnal which can be used to quit program
	memset(&sigacts, 0, sizeof(sigacts));
	sigacts.sa_handler = stop_loop_mode;
	if(sigaction(SIGINT, &sigacts, NULL) == -1) //Ctrl+C signal
		exit_error("Setting sigaction(SIGINT) failed");

	if(pcap_setfilter(gg_dt.pcap, &fprog) < 0)
		exit_error(pcap_geterr(gg_dt.pcap));

	if(gg_dt.read_fname[0] != 0) { //OFF LINE sniffing mode
	//NOTE: We have to use pcap_loop, when reading from file, because
	//      pcap_dispatch is not able to determine EOF.
		status = pcap_loop(gg_dt.pcap, -1, callback, NULL);
		if(status == -1){
			sprintf(errbuf, "pcap_loop: %s", pcap_geterr(gg_dt.pcap));
			exit_error(errbuf);}
	}
	else { //ON LINE sniffing mode (with user interface)
		pcap_fd = pcap_get_selectable_fd(gg_dt.pcap);

		// set terminal properties that we can read every single entered key
		if(tcgetattr(STDIN_FILENO, &old_opts) == -1)
			fprintf(stderr, "warning: terminal input settings can not be set\n");
		else{
			memcpy(&new_opts, &old_opts, sizeof(new_opts));
			new_opts.c_lflag &= ~(ICANON | ECHO);
			tcsetattr(STDIN_FILENO, TCSANOW, &new_opts);
			gg_dt.fl.term_ok = 1;}

		// selectable interaface... LOOP FOREVER :)
		for(gg_dt.fl.main_loop = 1; gg_dt.fl.main_loop;) {
		// NOTE: we CAN NOT exit program immediately from here because
		//       terminal will fucked up (only 'break' is safe)
			FD_ZERO(&rd_fds);
			FD_SET(pcap_fd, &rd_fds);
			FD_SET(fileno(stdin), &rd_fds);

			status = select(MAX(0, pcap_fd) + 1, &rd_fds, NULL, NULL, NULL);
			if(status == -1){
				if(errno == EINTR) continue; //NOTE: interrupt is not an error
				fprintf(stderr, "select: %s\n", strerror(errno));
				break;} //no exit-status error :/

			if(FD_ISSET(pcap_fd, &rd_fds)){
				if(pcap_dispatch(gg_dt.pcap, -1, callback, NULL) == -1){
					fprintf(stderr, "pcap_dispatch: %s", pcap_geterr(gg_dt.pcap));
					break;} //no exit-status error :/
				continue;}

			if(FD_ISSET(fileno(stdin), &rd_fds)){
				if(gg_dt.spooftool_input) //input redirection to spooftool callback
					gg_dt.spooftool_input = spooftool_input_callback();
				else
					switch(opt = tolower(getchar())){ //main input parsing
					case 'q':
						gg_dt.fl.main_loop = 0; break;
					case 'd':
						traficdb_dumpdb();
						printf("\n");
						break;
					case 'c': case 's':
						// parse input data in spooftool init :)
						gg_dt.spooftool_input = spooftool_input_init(opt);
						break;
					case ' ':
						toggle_visualization_mode();
						break;
					case 'h':
						printf("h\t\thow do you think, where are you now...\n"
"c\t\tsend spoofed packet to client\n"
"d\t\tprint captured client<->server connection(s)\n"
"q\t\tquit (Ctr+C also can be used)\n"
"s\t\tsend spoofed packet to server\n"
"space\t\tstart/stop visualization mode\n\n");
					}
			}
		}

		// restore the previous terminal settings
		if(gg_dt.fl.term_ok) tcsetattr(STDIN_FILENO, TCSANOW, &old_opts);
	}

	if(gg_dt.fl.show_conn_onexit) traficdb_dumpdb();

	clean_up();
	exit(EXIT_SUCCESS);
}

// Called on every sniffed packet by pcap_loop
void callback(u_char *args, const struct pcap_pkthdr *header,
	const u_char *packet)
{
	int status, db_stat, pack_type, len, full_date = 0;
	time_t tm_t;
	struct tm tm_p, tm_now;
 	struct ether_header *eth_h;
	struct iphdr *ip_h;
	struct tcphdr *tcp_h;
	uint16_t sprt, dprt;
	char sadr[20], dadr[20];
	u_char data_pack[1024*16];

	if((status = sniftool_is_gg_packet(packet)) > 0) {
		eth_h = (struct ether_header *)(packet);
		ip_h = (struct iphdr *)(eth_h + 1);
		tcp_h = (struct tcphdr *)(ip_h + 1);

		len = sizeof(data_pack);
		pack_type = sniftool_get_packet_type((u_char *)(ip_h), data_pack, &len);

		// it looks like empty or odd packet - just drop it
		// NOTE: we have to drop it before traficdb_updatedb because of DUP
		//       detection (if first packet is truncated then we will not
		//       pass complete one, which should come later)
		if(pack_type == 0) return;

		// update gg traffic (client2server connections)
		db_stat = traficdb_updatedb((status & 0x2) == 0x2, pack_type, eth_h,
				ip_h, (struct udphdr *)tcp_h, tcp_h->seq);

		// if traficdb detect that we've got DUP packet drop it
		if(db_stat == 3) return;

		// if partial packet try to complete all its data :)
		if(pack_type == -1)
			// if packet is not full yet drop parsing
			if(tcpipdata_add_partial_data(ip_h, (struct udphdr *)tcp_h,
					data_pack, &len, &pack_type) == 0) return;

		gg_dt.off_visual_packet_counter++;
		// dump packets to output only if visualization mode is on
		if(!gg_dt.fl.visualization) return;

		// packets to drop on default (communication stuff and so...)
		if(!gg_dt.fl.verbose && (pack_type == GG_LOGIN_OK ||
				pack_type == GG_LIST_EMPTY || pack_type == GG_SEND_MSG_ACK ||
				pack_type == GG_PING || pack_type == GG_PONG ||
				pack_type == GG_LOGIN80_OK || pack_type == GG_XML_EVENT))
			return;

		// status packets (drop if we don't need it)
		if(gg_dt.fl.drop_status && (pack_type == GG_NEW_STATUS80 ||
				pack_type == GG_NEW_STATUS || pack_type == GG_STATUS80 ||
				pack_type == GG_STATUS77 || pack_type == GG_STATUS60 ||
				pack_type == GG_STATUS))
			return;

		// drop local net traffic if we don't want it
		if(gg_dt.fl.drop_local)
			if(ip_h->saddr == gg_dt.localip || ip_h->daddr == gg_dt.localip)
				return;

		tm_t = time(NULL);
		gmtime_r(&tm_t, &tm_now);
		gmtime_r(&header->ts.tv_sec, &tm_p);
		if(tm_p.tm_year != tm_now.tm_year || tm_p.tm_mon != tm_now.tm_mon
				|| tm_p.tm_mday != tm_now.tm_mday)
			full_date = 1; //show full date if packet_time is not today

		sprt = ntohs(tcp_h->source);
		dprt = ntohs(tcp_h->dest);
		strcpy(sadr, inet_ntoa(*((struct in_addr *)(&ip_h->saddr))));
		strcpy(dadr, inet_ntoa(*((struct in_addr *)(&ip_h->daddr))));

		// print packet "header" info
		if(full_date)
			printf("%d-%.2d-%.2d ", tm_p.tm_year + 1900, tm_p.tm_mon, tm_p.tm_mday);
		printf("%.2d:%.2d.%.2d  (%s:%u --> %s:%u)\n",
				tm_p.tm_hour, tm_p.tm_min, tm_p.tm_sec, sadr, sprt, dadr, dprt);

		if(gg_dt.fl.verbose) //print some useful debugging info in verbose mode
			printf("pack_stat: 0x%.2x, pack_type: 0x%.2x (%s), msg_len: %u\n",
					status, pack_type, sniftool_get_strpacktype((status & 0x2) == 0x2,
					pack_type), len);

		if((status & 0x2) == 0x2) //client->server trafic
			switch(pack_type) {
			case GG_SEND_MSG80:
			case GG_SEND_MSG:
				sniftool_dump_sendmsg(data_pack, len, pack_type); break;
			case GG_NEW_STATUS80:
			case GG_NEW_STATUS:
				sniftool_dump_newstatus(data_pack, len, pack_type); break;
			case GG_ADD_NOTIFY:
			case GG_REMOVE_NOTIFY:
			case GG_NOTIFY_FIRST:
			case GG_NOTIFY_LAST:
				sniftool_dump_notify(data_pack, len, pack_type); break;
			case GG_LIST_EMPTY:
				printf("  LIST_EMPTY\n"); break;
			case GG_LOGIN80:
			case GG_LOGIN70:
			case GG_LOGIN60:
			case GG_LOGIN_EXT:
			case GG_LOGIN:
				sniftool_dump_login(data_pack, len, pack_type); break;
			}
		else //server->client traffic
			switch(pack_type) {
			case GG_RECV_MSG80:
			case GG_RECV_MSG:
				sniftool_dump_recvmsg(data_pack, len, pack_type); break;
			case GG_STATUS80:
			case GG_STATUS77:
			case GG_STATUS60:
			case GG_STATUS:
				sniftool_dump_status(data_pack, len, pack_type); break;
			case GG_NOTIFY_REPLY80:
			case GG_NOTIFY_REPLY77:
			case GG_NOTIFY_REPLY60:
			case GG_NOTIFY_REPLY:
				sniftool_dump_notifyreply(data_pack, len, pack_type); break;
			case GG_WELCOME:
				sniftool_dump_welcome(data_pack); break;
			}

		printf("\n");
		fflush(stdout);
	}
}

// Cyclically visualization state changing
void toggle_visualization_mode()
{
	gg_dt.fl.visualization ^= 1;
	printf("Visualization mode is: %s",
			gg_dt.fl.visualization ? "on" : "off");
	if(gg_dt.fl.visualization)
		printf(" | [%u packets detected in off-visual mode]",
				gg_dt.off_visual_packet_counter);
	else gg_dt.off_visual_packet_counter = 0;
		printf("\n\n");
}

// Convert cp1250 (GG codepage) to UTF-8 (only to std. ASCII)
char *cp1250_to_utf8_plchar(char *msg)
{
	int x, x_mat, str_len;
	char conv_matrix[][2] = {{0xa5, 'A'}, {0xc6, 'C'}, {0xca, 'E'}, {0xa3, 'L'},
{0xd1, 'N'}, {0xd3, 'O'}, {0x8c, 'S'}, {0xaf, 'Z'}, {0x8f, 'Z'}, {0xb9, 'a'},
{0xe6, 'c'}, {0xea, 'e'}, {0xb3, 'l'}, {0xf1, 'n'}, {0xf3, 'o'}, {0x9c, 's'},
{0xbf, 'z'}, {0x9f, 'z'}};

	str_len = strlen(msg);
	for(x = 0; x < str_len; x++)
		for(x_mat = 0; x_mat < sizeof(conv_matrix)/2; x_mat++)
			if(msg[x] == conv_matrix[x_mat][0])
				msg[x] = conv_matrix[x_mat][1];

	return msg;
}

// Get UIN alias from alias table (if not available return NULL)
char *get_uinalias(int uin)
{
	int x;

	for(x = 0; x < gg_dt.uinal_size; x++)
		if(gg_dt.uinal[x].uin == uin)
			return gg_dt.uinal[x].alias;
	return NULL;
}

// Load UIN alias table from file
int load_uinalias_table(const char *fname)
{
	int uin, aliases_nb = 50;
	FILE *fd;
	char rd_buffer[100]; //only one line is read so this size is enough
	char *ptr, *ptr2;
	void *realloc_ptr;

	if((fd = fopen(fname, "r")) == NULL) return -1;

	// as default alloc memory for 50 aliases (if it is not enough realloc)
	gg_dt.uinal = (struct aliases *)malloc(sizeof(struct aliases)*aliases_nb);
	if(gg_dt.uinal == NULL) return -1;

	for(gg_dt.uinal_size = 0;;) {
		// break reading on error or when EOF occurred
		if(fgets(rd_buffer, sizeof(rd_buffer), fd) == NULL) break;

		// skip lines with '#' at beginning
		if(rd_buffer[0] == '#') continue;

		if((uin = (int)strtol(rd_buffer, &ptr, 10)) == 0) continue;

		if(gg_dt.uinal_size == aliases_nb){ //realloc (add space for new aliases)
			aliases_nb += 50;
			realloc_ptr = (struct aliases *)realloc(gg_dt.uinal,
					sizeof(struct aliases)*aliases_nb);
			if(realloc_ptr == NULL) return -1;
			gg_dt.uinal = (struct aliases *)realloc_ptr;}

		// prepare alias string to copy (strip and cut if needed)
		while(isspace(*ptr)) ptr++;
		ptr2 = &ptr[strlen(ptr) - 1];
		while(isspace(*ptr2)) ptr2--;
		*&ptr2[1] = 0;
		if(strlen(ptr) > GGCAP_MAX_ALIAS_SIZE) ptr[GGCAP_MAX_ALIAS_SIZE] = 0;

		gg_dt.uinal[gg_dt.uinal_size].uin = uin;
		strcpy(gg_dt.uinal[gg_dt.uinal_size].alias, ptr);
		gg_dt.uinal_size++;
	}

	fclose(fd);
	return 0;
}

// Get network interface IP
uint32_t get_devip(const char *iface)
{
	int sock;
	struct ifreq ifr;

	bzero(&ifr, sizeof(ifr));
	strncpy(ifr.ifr_name, iface, sizeof(ifr.ifr_name));

	if((sock = socket(PF_INET, SOCK_STREAM, 0)) == -1) return -1;
	if((ioctl(sock, SIOCGIFADDR, &ifr)) == -1) return -1;
	close(sock);

	return *(uint32_t *)(&((struct sockaddr_in *)(&ifr.ifr_addr))->sin_addr);
}

// Dump memory as hext to buffer
// NOTE: sizeof(buf) >= 2*len + 1
char *strdumphex(void *mem, int len, char *buf)
{
	int x;

	if(len <= 0) return NULL;
	for(x = 0; x < len; x++)
		sprintf(&buf[2*x], "%.2X", ((char *)mem)[x]);
	buf[2*x - 1] = 0;
	return buf;
}

// Clean up allocated memory and opened descriptors
void clean_up()
{
	if(gg_dt.pcap) pcap_close(gg_dt.pcap);
	if(gg_dt.uinal) free(gg_dt.uinal);
	traficdb_clean_up();
	tcpipdata_collector_clean_up();
	spooftool_clean_up();
}

// Clean up, print err message and exit with EXIT_FAILURE status
void exit_error(const char *errmsg)
{
	clean_up();
	fprintf(stderr, "%s: %s\n", APP_NAME, errmsg);
	exit(EXIT_FAILURE);
}

void usage_help()
{
	printf("Usage: %s [options]\n"
"  -h, --help\t\t\tprint this help text and exit\n"
"  -V, --version\t\t\tprint program version and exit\n"
"  -i NIF, --iface=NIF\t\tuse NIF network interface (default: %s)\n"
"  -R FILE, --read=FILE\t\toff-line sniffing (use pcap dump file)\n"
"  -f BFC, --filter=BFC\t\tuse BFC pcapfilter expression\n"
"  -a FILE, --aliases=FILE\tload UIN alias-table from FILE\n"
"  --vismod-off\t\t\tdo not go into visualization mode at start\n"
"  --nostatus\t\t\tdo not dump status change packets\n"
"  --nolocal\t\t\thide local host network traffic\n"
"  --show-conn-onexit\t\tprint captured client<->server connection(s)\n"
"  -v, --verbose\t\t\tverbose output (print all we've got)\n",
	APP_NAME, gg_dt.iface);

#ifdef GGCAP_MORALITY_NOTE
	char moral_quote[] = "Morality, like art, means drawing a line someplace.";
	char moral_author[] = "Oscar Wilde";

	printf("\nATTENTION:\n\"%s\"\n%*s\n", moral_quote,
			(int)sizeof(moral_quote), moral_author);

	printf("\nPlease, do NOT abuse this software!\n\n");
#endif
}
