/*
    ggcap - traficdb.c
    Copyright (c) 2009 Arkadiusz Bokowy
*/

#include <string.h>
#include <arpa/inet.h>

#include "ggcap.h"

// Dump client2server connection database to stdout
void traficdb_dumpdb()
{
	struct client2server_info *db_info;
	time_t dif_time;
	char cliadr[20], srvadr[20];
	int x;

	if(gg_dt.cl2srv_db->len == 0)
		printf("There is no entry in connections database.\n");
	else
		printf("No.| last | type |         client        |         server        |\n");

	for(x = 0; x < gg_dt.cl2srv_db->len; x++){
		db_info = &g_array_index(gg_dt.cl2srv_db, struct client2server_info, x);

		dif_time = time(NULL) - db_info->last_seen;
		strcpy(cliadr, inet_ntoa(*((struct in_addr *)(&db_info->cliaddr))));
		strcpy(srvadr, inet_ntoa(*((struct in_addr *)(&db_info->srvaddr))));

		printf("%2d | %4u |  %.2x  | %15s:%-5u | %15s:%-5u |\n",
				x + 1, (int)dif_time, db_info->type, cliadr, db_info->cliprt,
				srvadr, db_info->srvprt);
	}
}

// Update (if necessery) client2server connection database
// NOTE: Function return values:
// 0 - entry is in db,
// 1 - added new entry,
// 2 - updated entry,
// 3 - duplicated packet
int traficdb_updatedb(int to_server, int pack_type, const struct ether_header *eth_h,
	const struct iphdr *ip_h, const struct udphdr *tcp_udp_h, uint32_t tcp_seq)
{
	struct client2server_info cl2srv_inf, *db_inf;
	int x;

	// create db entry from received data
	cl2srv_inf.last_seen = time(NULL);
	cl2srv_inf.tcp_seq = tcp_seq;
	cl2srv_inf.type = traficdb_get_client_type(pack_type);
	traficdb_client_server_detect(&cl2srv_inf, to_server, eth_h, ip_h, tcp_udp_h);

	// return DUP info if new tcp seq == last tcp seq
	if(traficdb_get_last_seq(&cl2srv_inf) == tcp_seq) return 3;

	// scan if created entry is already in db
	for(x = 0; x < gg_dt.cl2srv_db->len; x++){
		db_inf = &g_array_index(gg_dt.cl2srv_db, struct client2server_info, x);

		if(memcmp(&cl2srv_inf, db_inf, sizeof(cl2srv_inf)) == 0)
			return 0;

		// if client<->server ip:port match data stored in db but something
		// else not (see memcmp above) -> update db
		if(cl2srv_inf.cliaddr == db_inf->cliaddr &&
				cl2srv_inf.cliprt == db_inf->cliprt &&
				cl2srv_inf.srvaddr == db_inf->srvaddr &&
				cl2srv_inf.srvprt == db_inf->srvprt) {

			// if new client type == 0, then do NOT change this value in db
			if(cl2srv_inf.type == 0) cl2srv_inf.type = db_inf->type;

			memcpy(db_inf, &cl2srv_inf, sizeof(cl2srv_inf));
			return 2;
		}
	}

	g_array_append_val(gg_dt.cl2srv_db, cl2srv_inf);
	return 1;
}

// Return unique tcp seq number or 0 if connection is not in db
// NOTE: this function uses only src and dst ip:port from client2server_info
int traficdb_get_last_seq(const struct client2server_info *cl2srv_inf)
{
	struct client2server_info *db_inf;
	int x;

	for(x = 0; x < gg_dt.cl2srv_db->len; x++){
		db_inf = &g_array_index(gg_dt.cl2srv_db, struct client2server_info, x);

		// find client<->server ip:port match
		if(cl2srv_inf->cliaddr == db_inf->cliaddr &&
				cl2srv_inf->cliprt == db_inf->cliprt &&
				cl2srv_inf->srvaddr == db_inf->srvaddr &&
				cl2srv_inf->srvprt == db_inf->srvprt)
			return db_inf->tcp_seq;
	}
	return 0;
}

// Fill in client<->server ip:port field in client2server_info
void traficdb_client_server_detect(struct client2server_info *cl2srv_inf,
	int to_server, const struct ether_header *eth_h, const struct iphdr *ip_h,
	const struct udphdr *tcp_udp_h)
{
	if(to_server){
		memcpy(cl2srv_inf->climac, eth_h->ether_shost, sizeof(cl2srv_inf->climac));
		memcpy(cl2srv_inf->srvmac, eth_h->ether_dhost, sizeof(cl2srv_inf->srvmac));
		cl2srv_inf->cliaddr = ip_h->saddr;
		cl2srv_inf->srvaddr = ip_h->daddr;
		cl2srv_inf->cliprt = ntohs(tcp_udp_h->source);
		cl2srv_inf->srvprt = ntohs(tcp_udp_h->dest);}
	else{
		memcpy(cl2srv_inf->climac, eth_h->ether_dhost, sizeof(cl2srv_inf->climac));
		memcpy(cl2srv_inf->srvmac, eth_h->ether_shost, sizeof(cl2srv_inf->srvmac));
		cl2srv_inf->cliaddr = ip_h->daddr;
		cl2srv_inf->srvaddr = ip_h->saddr;
		cl2srv_inf->cliprt = ntohs(tcp_udp_h->dest);
		cl2srv_inf->srvprt = ntohs(tcp_udp_h->source);}
}

// Determine client type (version) form packet ID
int traficdb_get_client_type(int pack_type)
{
	switch(pack_type)	{
	case GG_SEND_MSG80:
	case GG_RECV_MSG80:
	case GG_NEW_STATUS80:
	case GG_STATUS80:
	case GG_NOTIFY_REPLY80:
	case GG_LOGIN80:
		return 0x80;
	case GG_STATUS77:
	case GG_NOTIFY_REPLY77:
		return 0x77;
	case GG_LOGIN70:
		return 0x70;
	case GG_STATUS60:
	case GG_NOTIFY_REPLY60:
	case GG_LOGIN60:
		return 0x60;
	}

	// return 0 means that we don't know what client version it is
	return 0;
}

void traficdb_clean_up()
{
	// destroy database array (free elements too)
	if(gg_dt.cl2srv_db) g_array_free(gg_dt.cl2srv_db, TRUE);
}

void traficdb_initialize()
{
	// create database array (preload 10 entry)
	gg_dt.cl2srv_db = g_array_sized_new(0, 0,
			sizeof(struct client2server_info), 10);
}